@@ -127,7 +127,11 @@ static int PylibMC_Client_init(PylibMC_Client *self, PyObject *args,
 #endif
     }
 
-    memcached_behavior_set(self->mc, MEMCACHED_BEHAVIOR_BINARY_PROTOCOL, bin);
+    rc = memcached_behavior_set(self->mc, MEMCACHED_BEHAVIOR_BINARY_PROTOCOL, bin);
+    if (rc != MEMCACHED_SUCCESS) {
+        PyErr_SetString(PyExc_RuntimeError, "binary protocol behavior set failed");
+        goto error;
+    }
 
     while ((c_srv = PyIter_Next(srvs_it)) != NULL) {
         unsigned char stype;
@@ -159,8 +163,11 @@ static int PylibMC_Client_init(PylibMC_Client *self, PyObject *args,
             } else {
                 set_stype = stype;
                 if (stype == PYLIBMC_SERVER_UDP) {
-                    memcached_behavior_set(self->mc,
-                        MEMCACHED_BEHAVIOR_USE_UDP, 1);
+                    rc = memcached_behavior_set(self->mc, MEMCACHED_BEHAVIOR_USE_UDP, 1);
+                    if (rc != MEMCACHED_SUCCESS) {
+                        PyErr_SetString(PyExc_RuntimeError, "udp behavior set failed");
+                        goto it_error;
+                    }
                 }
             }
 
@@ -471,7 +478,7 @@ static PyObject *PylibMC_Client_get(PylibMC_Client *self, PyObject *arg) {
 static PyObject *PylibMC_Client_gets(PylibMC_Client *self, PyObject *arg) {
     const char* keys[2];
     size_t keylengths[2];
-    memcached_result_st *res, *res0;
+    memcached_result_st *res = NULL;
     memcached_return rc;
     PyObject* ret = NULL;
 
@@ -494,12 +501,8 @@ static PyObject *PylibMC_Client_gets(PylibMC_Client *self, PyObject *arg) {
     Py_BEGIN_ALLOW_THREADS;
 
     rc = memcached_mget(self->mc, keys, keylengths, 1);
-    if (rc == MEMCACHED_SUCCESS) {
-        res0 = memcached_result_create(self->mc, NULL);
-        /* this will be NULL if the key wasn't found, or
-           memcached_result_st if it was */
-        res = memcached_fetch_result(self->mc, res0, &rc);
-    }
+    if (rc == MEMCACHED_SUCCESS)
+        res = memcached_fetch_result(self->mc, res, &rc);
 
     Py_END_ALLOW_THREADS;
 
@@ -509,8 +512,7 @@ static PyObject *PylibMC_Client_gets(PylibMC_Client *self, PyObject *arg) {
                             memcached_result_cas(res));
 
         /* we have to fetch the last result from the mget cursor */
-        res = memcached_fetch_result(self->mc, res0, &rc);
-        if (res || rc != MEMCACHED_END) {
+        if (NULL != memcached_fetch_result(self->mc, NULL, &rc)) {
             memcached_quit(self->mc);
             Py_DECREF(ret);
             ret = NULL;
@@ -523,10 +525,8 @@ static PyObject *PylibMC_Client_gets(PylibMC_Client *self, PyObject *arg) {
         ret = PylibMC_ErrFromMemcached(self, "memcached_gets", rc);
     }
 
-    /* deallocate any indirect buffers, even though the struct itself
-       is on our stack */
-    if (res0 != NULL) {
-        memcached_result_free(res0);
+    if (res != NULL) {
+        memcached_result_free(res);
     }
 
     return ret;
@@ -886,7 +886,7 @@ static bool _PylibMC_RunSetCommand(PylibMC_Client* self,
         pylibmc_mset *mset = &msets[pos];
 
         char *value = mset->value;
-        size_t value_len = mset->value_len;
+        size_t value_len = (size_t)mset->value_len;
         uint32_t flags = mset->flags;
 
 #ifdef USE_ZLIB
@@ -998,30 +998,26 @@ static PyObject *PylibMC_Client_cas(PylibMC_Client *self, PyObject *args,
 }
 
 static PyObject *PylibMC_Client_delete(PylibMC_Client *self, PyObject *args) {
-    PyObject *retval;
     char *key;
-    Py_ssize_t key_sz;
+    Py_ssize_t key_len = 0;
     memcached_return rc;
 
-    retval = NULL;
-    if (PyArg_ParseTuple(args, "s#:delete", &key, &key_sz)
-            && _PylibMC_CheckKeyStringAndSize(key, key_sz)) {
+    if (PyArg_ParseTuple(args, "s#:delete", &key, &key_len)
+            && _PylibMC_CheckKeyStringAndSize(key, key_len)) {
         Py_BEGIN_ALLOW_THREADS;
-        rc = memcached_delete(self->mc, key, key_sz, 0);
+        rc = memcached_delete(self->mc, key, key_len, 0);
         Py_END_ALLOW_THREADS;
         switch (rc) {
             case MEMCACHED_SUCCESS:
                 Py_RETURN_TRUE;
-                break;
             case MEMCACHED_FAILURE:
             case MEMCACHED_NOTFOUND:
             case MEMCACHED_NO_KEY_PROVIDED:
             case MEMCACHED_BAD_KEY_PROVIDED:
                 Py_RETURN_FALSE;
-                break;
             default:
                 return PylibMC_ErrFromMemcachedWithKey(self, "memcached_delete",
-                                                       rc, key, key_sz);
+                                                       rc, key, key_len);
         }
     }
 
@@ -1033,15 +1029,25 @@ static PyObject *_PylibMC_IncrSingle(PylibMC_Client *self,
                                      _PylibMC_IncrCommand incr_func,
                                      PyObject *args) {
     char *key;
-    Py_ssize_t key_len;
-    unsigned int delta = 1;
+    Py_ssize_t key_len = 0;
+    int delta = 1;
 
-    if (!PyArg_ParseTuple(args, "s#|I", &key, &key_len, &delta)) {
+    if (!PyArg_ParseTuple(args, "s#|i", &key, &key_len, &delta)) {
         return NULL;
     } else if (!_PylibMC_CheckKeyStringAndSize(key, key_len)) {
         return NULL;
     }
 
+    if (delta < 0L) {
+        PyErr_SetString(PyExc_ValueError, "delta must be positive");
+        return NULL;
+    }
+
+    if ((unsigned int)delta != delta) {
+        PyErr_Format(PyExc_OverflowError, "%d", delta);
+        return NULL;
+    }
+
     pylibmc_incr incr = { key, key_len,
                           incr_func, delta,
                           0 };
@@ -1060,132 +1066,95 @@ static PyObject *_PylibMC_IncrSingle(PylibMC_Client *self,
 static PyObject *_PylibMC_IncrMulti(PylibMC_Client *self,
                                     _PylibMC_IncrCommand incr_func,
                                     PyObject *args, PyObject *kwds) {
-  /**
-   * Takes an iterable of keys and a single delta (that defaults to 1)
-   * to be applied to all keys. Consider the return value and exception
-   * behaviour to be undocumented, for now it returns None and throws an
-   * exception on an error incrementing any key
-   */
-
+    PyObject *key = NULL;
     PyObject *keys = NULL;
+    PyObject *keys_tmp = NULL;
     PyObject *key_prefix = NULL;
-    PyObject *prefixed_keys = NULL;
     PyObject *retval = NULL;
     PyObject *iterator = NULL;
     unsigned int delta = 1;
+    size_t nkeys = 0, i = 0;
 
     static char *kws[] = { "keys", "key_prefix", "delta", NULL };
 
     if (!PyArg_ParseTupleAndKeywords(args, kwds, "O|SI", kws,
-                                     &keys, &key_prefix, &delta)) {
-        return NULL;
-    }
-
-    size_t nkeys = (size_t)PySequence_Size(keys);
-    if(nkeys == -1)
+                                     &keys, &key_prefix, &delta))
         return NULL;
 
-    if(key_prefix == NULL || PyString_Size(key_prefix) < 1) {
-        /* if it's 0-length, we can safely pretend it doesn't exist */
-        key_prefix = NULL;
-    }
-    if(key_prefix != NULL) {
-        if(!_PylibMC_CheckKey(key_prefix)) {
-            return NULL;
-        }
+    nkeys = (size_t)PySequence_Size(keys);
+    if (nkeys == -1)
+        return NULL;
 
-        prefixed_keys = PyList_New(nkeys);
-        if(prefixed_keys == NULL) {
+    if (key_prefix != NULL) {
+        if (!_PylibMC_CheckKey(key_prefix))
             return NULL;
-        }
+
+        if (PyString_Size(key_prefix) == 0)
+            key_prefix = NULL;
     }
 
-    pylibmc_incr* incrs = PyMem_New(pylibmc_incr, nkeys);
-    if(incrs == NULL) {
+    keys_tmp = PyList_New(nkeys);
+    if (keys_tmp == NULL)
+        return NULL;
+
+    pylibmc_incr *incrs = PyMem_New(pylibmc_incr, nkeys);
+    if (incrs == NULL)
         goto cleanup;
-    }
 
     iterator = PyObject_GetIter(keys);
-    if(iterator == NULL) {
+    if (iterator == NULL)
         goto cleanup;
-    }
 
-    PyObject *key = NULL;
-    size_t idx = 0;
+    /* Build pylibmc_incr structs, prefixed as appropriate. */
+    for (i = 0; (key = PyIter_Next(iterator)) != NULL; i++) {
+        pylibmc_incr *incr = incrs + i;
 
-    /**
-     * Build our list of keys, prefixed as appropriate, and turn that
-     * into a list of pylibmc_incr objects that can be incred in one
-     * go. We're not going to own references to the prefixed keys: so
-     * that we can free them all at once, we'll give ownership to a list
-     * of them (prefixed_keys) which we'll DECR once at the end
-     */
+        if (!_PylibMC_CheckKey(key))
+            goto loopcleanup;
 
-    while((key = PyIter_Next(iterator)) != NULL) {
-        if(!_PylibMC_CheckKey(key)) goto loopcleanup;
-
-        if(key_prefix != NULL) {
-            PyObject* newkey = NULL;
-
-            newkey = PyString_FromFormat("%s%s",
-                                         PyString_AS_STRING(key_prefix),
-                                         PyString_AS_STRING(key));
-
-            if(!_PylibMC_CheckKey(newkey)) {
-                Py_XDECREF(newkey);
-                goto loopcleanup;
-            }
-
-            /* steals our reference */
-            if(PyList_SetItem(prefixed_keys, idx, newkey) == -1) {
-                /* it wasn't stolen before the error */
-                Py_DECREF(newkey);
-                goto loopcleanup;
-            }
-
-            /* the list stole our reference to it, and we're going to rely
-               on that list to maintain it while we release the GIL, but
-               since we DECREF the key in loopcleanup we need to INCREF it
-               here */
+        /* prefix `key` with `key_prefix` */
+        if (key_prefix != NULL) {
+            PyObject* newkey = PyString_FromFormat("%s%s",
+                                                   PyString_AS_STRING(key_prefix),
+                                                   PyString_AS_STRING(key));
             Py_DECREF(key);
-            Py_INCREF(newkey);
             key = newkey;
         }
 
-        if(PyString_AsStringAndSize(key, &incrs[idx].key,
-                                    &incrs[idx].key_len) == -1)
+        Py_INCREF(key);
+        if (PyList_SetItem(keys_tmp, i, key) == -1)
             goto loopcleanup;
 
-        incrs[idx].delta = delta;
-        incrs[idx].incr_func = incr_func;
+        /* Populate pylibmc_incr */
+        if (PyString_AsStringAndSize(key, &incr->key, &incr->key_len) == -1)
+            goto loopcleanup;
+        incr->delta = delta;
+        incr->incr_func = incr_func;
         /* After incring we have no way of knowing whether the real result is 0
          * or if the incr wasn't successful (or if noreply is set), but since
          * we're not actually returning the result that's okay for now */
-        incrs[idx].result = 0;
+        incr->result = 0;
 
 loopcleanup:
         Py_DECREF(key);
-
-        if(PyErr_Occurred())
-            break;
-
-        idx++;
+        if (PyErr_Occurred())
+            goto cleanup;
     } /* end each key */
 
-    /* iteration error */
-    if (PyErr_Occurred()) goto cleanup;
-
     _PylibMC_IncrDecr(self, incrs, nkeys);
 
-    /* if that failed, there's an exception on the stack */
-    if(PyErr_Occurred()) goto cleanup;
-
-    retval = Py_None;
-    Py_INCREF(retval);
+    if (!PyErr_Occurred()) {
+        retval = Py_None;
+        Py_INCREF(retval);
+    } else {
+        Py_XDECREF(retval);
+        retval = NULL;
+    }
 
 cleanup:
-    PyMem_Free(incrs);
-    Py_XDECREF(prefixed_keys);
+    if (incrs != NULL)
+        PyMem_Free(incrs);
+    Py_DECREF(keys_tmp);
     Py_XDECREF(iterator);
 
     return retval;
@@ -1206,34 +1175,42 @@ static PyObject *PylibMC_Client_incr_multi(PylibMC_Client *self, PyObject *args,
 
 static bool _PylibMC_IncrDecr(PylibMC_Client *self,
                               pylibmc_incr *incrs, size_t nkeys) {
-    bool error = false;
     memcached_return rc = MEMCACHED_SUCCESS;
     _PylibMC_IncrCommand f = NULL;
-    size_t i;
+    size_t i, notfound = 0, errors = 0;
 
     Py_BEGIN_ALLOW_THREADS;
-    for (i = 0; i < nkeys && !error; i++) {
+    for (i = 0; i < nkeys; i++) {
         pylibmc_incr *incr = &incrs[i];
         uint64_t result = 0;
 
         f = incr->incr_func;
         rc = f(self->mc, incr->key, incr->key_len, incr->delta, &result);
+        /* TODO Signal errors through `incr` */
         if (rc == MEMCACHED_SUCCESS) {
             incr->result = result;
+        } else if (rc == MEMCACHED_NOTFOUND) {
+            notfound++;
         } else {
-            error = true;
+            errors++;
         }
     }
     Py_END_ALLOW_THREADS;
 
-    if (error) {
-        char *fname = (f == memcached_decrement) ? "memcached_decrement"
-                                               : "memcached_increment";
-        PylibMC_ErrFromMemcached(self, fname, rc);
-        return false;
-    } else {
-        return true;
+    if (errors + notfound) {
+        PyObject *exc = PylibMCExc_MemcachedError;
+
+        if (errors == 0)
+            exc = _exc_by_rc(MEMCACHED_NOTFOUND);
+        else if (errors == 1)
+            exc = _exc_by_rc(rc);
+
+        PyErr_Format(exc, "%d keys %s",
+                     (int)(notfound + errors),
+                     errors ? "failed" : "not found");
     }
+
+    return 0 == (errors + notfound);
 }
 /* }}} */
 
@@ -1363,8 +1340,10 @@ static PyObject *PylibMC_Client_get_multi(
         key_lens[i] = (size_t)(key_len + prefix_len);
 
         /* Skip empty keys */
-        if (!key_lens[i])
+        if (!key_lens[i]) {
+            Py_DECREF(ckey);
             continue;
+        }
 
         /* determine rkey, the prefixed ckey */
         if (prefix != NULL) {
@@ -1372,6 +1351,7 @@ static PyObject *PylibMC_Client_get_multi(
             PyString_Concat(&rkey, ckey);
             if (rkey == NULL)
                 goto earlybird;
+            Py_DECREF(rkey);
             rkey = PyString_FromFormat("%s%s",
                     prefix, PyString_AS_STRING(ckey));
         } else {
@@ -1420,9 +1400,7 @@ static PyObject *PylibMC_Client_get_multi(
 
         res = results + i;
 
-        /* Explicitly construct a key string object so that NUL-bytes and the
-         * likes can be contained within the keys (this is possible in the
-         * binary protocol.) */
+        /* Long-winded, but this way we can handle NUL-bytes in keys. */
         key_obj = PyString_FromStringAndSize(memcached_result_key_value(res) + prefix_len,
                                              memcached_result_key_length(res) - prefix_len);
         if (key_obj == NULL)
@@ -1443,8 +1421,9 @@ static PyObject *PylibMC_Client_get_multi(
         continue;
 
 unpack_error:
-            Py_DECREF(retval);
-            break;
+        Py_DECREF(retval);
+        retval = NULL;
+        break;
     }
 
 earlybird:
@@ -1790,11 +1769,6 @@ static PyObject *PylibMC_Client_get_stats(PylibMC_Client *self, PyObject *args)
     context.servers = NULL;  /* DEPRECATED */
     context.index = 0;
 
-    /* TODO Remove BC for 0.38 */
-#ifndef LIBMEMCACHED_VERSION_HEX
-#  define LIBMEMCACHED_VERSION_HEX 0x0
-#endif
-
 #if LIBMEMCACHED_VERSION_HEX >= 0x00038000
     memcached_server_function callbacks[] = {
         (memcached_server_function)_PylibMC_AddServerCallback
@@ -1876,26 +1850,21 @@ static PyObject *PylibMC_Client_clone(PylibMC_Client *self) {
 }
 /* }}} */
 
+static PyObject *_exc_by_rc(memcached_return rc) {
+    PylibMC_McErr *err;
+    for (err = PylibMCExc_mc_errs; err->name != NULL; err++)
+        if (err->rc == rc)
+            return err->exc;
+    return (PyObject *)PylibMCExc_MemcachedError;
+}
+
 static char *_get_lead(memcached_st *mc, char *buf, int n, const char *what,
         memcached_return error, const char *key, Py_ssize_t len) {
-    int sz = snprintf(buf, n, "error %d from %s", error, what);
+    int sz = snprintf(buf, n, "error %d from %.32s", error, what);
 
-    /*
-     * Need to protect from libmemcached versions as their
-     * memcached_server_instance_st are called (memcached_server_st *).
-     */
-#if LIBMEMCACHED_VERSION_HEX >= 0x00038000
-    if (key != NULL) {
-        memcached_return rc = MEMCACHED_FAILURE;
-        memcached_server_instance_st svr = NULL;
-
-        svr = memcached_server_by_key(mc, key, len, &rc);
-        if (svr && rc == MEMCACHED_SUCCESS) {
-            sz += snprintf(buf + sz, n - sz, " on %s:%d",
-                           svr->hostname, svr->port);
-        }
+    if (key != NULL && len) {
+        sz += snprintf(buf+sz, n-sz, "(%.32s)", key);
     }
-#endif
 
     return buf;
 }
@@ -1907,17 +1876,12 @@ static void _set_error(memcached_st *mc, memcached_return error, char *lead) {
     } else if (error == MEMCACHED_SUCCESS) {
         PyErr_Format(PyExc_RuntimeError, "error == MEMCACHED_SUCCESS");
     } else {
-        PylibMC_McErr *err;
-        PyObject *exc = (PyObject *)PylibMCExc_MemcachedError;
-
-        for (err = PylibMCExc_mc_errs; err->name != NULL; err++) {
-            if (err->rc == error) {
-                exc = err->exc;
-                break;
-            }
-        }
-
-        PyErr_Format(exc, "%s: %s", lead, memcached_strerror(mc, error));
+        PyObject *exc = _exc_by_rc(error);
+#if LIBMEMCACHED_VERSION_HEX >= 0x00049000
+        PyErr_Format(exc, "%s: %.200s", lead, memcached_last_error_message(mc));
+#else
+        PyErr_Format(exc, "%s: %.200s", lead, memcached_strerror(mc, error));
+#endif
     }
 }
 
@@ -2060,7 +2024,7 @@ static int _check_libmemcached_version(void) {
     uint8_t maj, min;
     char *ver, *dot, *tmp;
 
-    ver = dot = strndup(LIBMEMCACHED_VERSION_STRING, 8);
+    ver = dot = strdup(LIBMEMCACHED_VERSION_STRING);
     while ((tmp = strrchr(ver, '.')) != NULL) {
         dot = tmp;
         *dot = 0;
