@@ -41,6 +41,10 @@
 #include <Python.h>
 #include <libmemcached/memcached.h>
 
+#ifndef LIBMEMCACHED_VERSION_HEX
+#  define LIBMEMCACHED_VERSION_HEX 0x0
+#endif
+
 #include "pylibmc-version.h"
 
 /* Py_ssize_t appeared in Python 2.5. */
@@ -73,18 +77,20 @@ typedef memcached_return (*_PylibMC_SetCommand)(memcached_st *, const char *,
 typedef memcached_return (*_PylibMC_IncrCommand)(memcached_st *,
         const char *, size_t, unsigned int, uint64_t*);
 
+static PyObject *_exc_by_rc(memcached_return);
+
 typedef struct {
   char *key;
   Py_ssize_t key_len;
-  char* value;
+  char *value;
   Py_ssize_t value_len;
   time_t time;
   uint32_t flags;
 
   /* the objects that must be freed after the mset is executed */
-  PyObject* key_obj;
-  PyObject* prefixed_key_obj;
-  PyObject* value_obj;
+  PyObject *key_obj;
+  PyObject *prefixed_key_obj;
+  PyObject *value_obj;
 
   /* the success of executing the mset afterwards */
   int success;
@@ -151,7 +157,10 @@ static PylibMC_McErr PylibMCExc_mc_errs[] = {
     //{ MEMCACHED_BUFFERED, "Buffer, NULL },
     { MEMCACHED_BAD_KEY_PROVIDED, "BadKeyProvided", NULL },
     { MEMCACHED_INVALID_HOST_PROTOCOL, "InvalidHostProtocolError", NULL },
-    //{ MEMCACHED_SERVER_MARKED_DEAD,
+    { MEMCACHED_SERVER_MARKED_DEAD, "ServerDead", NULL },
+#ifdef MEMCACHED_SERVER_TEMPORARILY_DISABLED
+    { MEMCACHED_SERVER_TEMPORARILY_DISABLED, "ServerDown", NULL },
+#endif
     { MEMCACHED_UNKNOWN_STAT_KEY, "UnknownStatKey", NULL },
     //{ MEMCACHED_E2BIG, "TooBigError", NULL },
     { 0, NULL, NULL }
@@ -180,6 +189,7 @@ static PylibMC_Behavior PylibMC_behaviors[] = {
     { MEMCACHED_BEHAVIOR_RCV_TIMEOUT, "receive_timeout" },
     { MEMCACHED_BEHAVIOR_NUMBER_OF_REPLICAS, "num_replicas" },
     { MEMCACHED_BEHAVIOR_AUTO_EJECT_HOSTS, "auto_eject" },
+    { MEMCACHED_BEHAVIOR_RETRY_TIMEOUT, "retry_timeout" },
 #if LIBMEMCACHED_VERSION_HEX >= 0x00049000
     { MEMCACHED_BEHAVIOR_REMOVE_FAILED_SERVERS, "remove_failed" },
 #endif
@@ -193,7 +203,6 @@ static PylibMC_Behavior PylibMC_behaviors[] = {
     { MEMCACHED_BEHAVIOR_HASH_WITH_PREFIX_KEY, "_hash_with_prefix_key" },
     { MEMCACHED_BEHAVIOR_NOREPLY, "_noreply" },
     { MEMCACHED_BEHAVIOR_SORT_HOSTS, "_sort_hosts" },
-    { MEMCACHED_BEHAVIOR_RETRY_TIMEOUT, "_retry_timeout" },
     { MEMCACHED_BEHAVIOR_POLL_TIMEOUT, "_poll_timeout" },
     { MEMCACHED_BEHAVIOR_SOCKET_SEND_SIZE, "_socket_send_size" },
     { MEMCACHED_BEHAVIOR_SOCKET_RECV_SIZE, "_socket_recv_size" },
@@ -265,23 +274,23 @@ static PyObject *_PylibMC_Unpickle(const char *, size_t);
 static PyObject *_PylibMC_Pickle(PyObject *);
 static int _PylibMC_CheckKey(PyObject *);
 static int _PylibMC_CheckKeyStringAndSize(char *, Py_ssize_t);
-static int _PylibMC_SerializeValue(PyObject* key_obj,
-                                   PyObject* key_prefix,
-                                   PyObject* value_obj,
+static int _PylibMC_SerializeValue(PyObject *key_obj,
+                                   PyObject *key_prefix,
+                                   PyObject *value_obj,
                                    time_t time,
-                                   pylibmc_mset* serialized);
+                                   pylibmc_mset *serialized);
 static void _PylibMC_FreeMset(pylibmc_mset*);
 static PyObject *_PylibMC_RunSetCommandSingle(PylibMC_Client *self,
         _PylibMC_SetCommand f, char *fname, PyObject *args, PyObject *kwds);
-static PyObject *_PylibMC_RunSetCommandMulti(PylibMC_Client* self,
+static PyObject *_PylibMC_RunSetCommandMulti(PylibMC_Client *self,
         _PylibMC_SetCommand f, char *fname, PyObject *args, PyObject *kwds);
-static bool _PylibMC_RunSetCommand(PylibMC_Client* self,
+static bool _PylibMC_RunSetCommand(PylibMC_Client *self,
                                    _PylibMC_SetCommand f, char *fname,
-                                   pylibmc_mset* msets, size_t nkeys,
+                                   pylibmc_mset *msets, size_t nkeys,
                                    size_t min_compress);
-static int _PylibMC_Deflate(char* value, size_t value_len,
-                            char** result, size_t *result_len);
-static bool _PylibMC_IncrDecr(PylibMC_Client*, pylibmc_incr*, size_t);
+static int _PylibMC_Deflate(char *value, size_t value_len,
+                            char **result, size_t *result_len);
+static bool _PylibMC_IncrDecr(PylibMC_Client *, pylibmc_incr *, size_t);
 
 /* }}} */
 
